import numpy as np
from sympy import *
from sympy.physics.quantum import TensorProduct
from functools import reduce
from qiskit.quantum_info import *
from scipy.optimize import minimize
from numpy.testing import assert_array_almost_equal as aae


def view(mat, rounding = 10):
    display(Matrix(np.round(mat, rounding)))

class Mode:
    representation = 'numpy'

    @classmethod
    def toggle(cls):
        cls.representation = {'numpy': 'sympy',
                              'sympy': 'numpy'}[cls.representation]
        print(f'\033[1m{cls.representation.capitalize()}\033[0m mode activated')

    @classmethod
    def now(cls):
        print(f'\033[1m{cls.representation.capitalize()}\033[0m mode is \033[1mon\033[0m')



from FrozenYoghourt.mode import *
from FrozenYoghourt.maths import *


class Quantum:

    def random_local_gates(num_qubits):
        if Mode.representation == 'numpy':
            return Maths.tp(*[random_unitary(2) for i in range(num_qubits)])
        else:
            return Maths.tp(*[Gates.U(*symbols(f'theta_{i}, phi_{i}, lambda_{i}')) for i in range(num_qubits)])

    def epsilon(psi, num_qubits=2):
        if Mode.representation == 'numpy':
            return np.abs(psi.T @ Maths.tp(Gates.Y(), no_times=num_qubits) @ psi)[0]
        else:
            return Abs(psi.T @ Maths.tp(Gates.Y(), no_times=num_qubits) @ psi)[0]

    def global_phase(A, B):
        D = np.diag(np.conj(A).T @ B)

        if np.isclose(np.min(D), np.max(D)):
            return D[0]
        else:
            print('A and B are not equivalent up to global phase')
            return False
        
    def kron_decomp(U:np.ndarray):
        m = U.reshape(2, 2, 2, 2).transpose(0, 2, 1, 3).reshape(4, 4)

        u, sv, vh = np.linalg.svd(m)

        a = np.sqrt(sv[0]) * u[:, 0].reshape(2, 2)
        b = np.sqrt(sv[0]) * vh[0, :].reshape(2, 2)

        return a, b
    
   
    def gamma(U):
        n = int(np.log2(U.shape[0]))

        E = Maths.tp(Gates.Y(), no_times=n)
        return U @ E @ U.T @ E

    def chi(M, variable='x'):
        dim = M.shape[0]

        coef = np.array([1])
        Mk = np.array(M)

        for k in range(1, dim + 1):
            ak = -Mk.trace() / k
            coef = np.append(coef, ak)
            Mk += np.diag(np.repeat(ak, dim))
            Mk = np.dot(M, Mk)

        if Mode.representation == 'numpy':
            return coef

        else:
            var = symbols(variable)
            variable_mat = Matrix([var ** n for n in range(dim + 1)])
            coef = Matrix(coef).T
            char_poly = simplify(coef @ variable_mat)[0]
            return char_poly

    def double_cosets(g, h):
        E = Gates.Magic()

        g, h = Quantum.to_su(g), Quantum.to_su(h)

        u, v = np.conj(E).T @ g @ E, np.conj(E).T @ h @ E

        D1, P1 = np.linalg.eig(u @ u.T)
        D2, P2 = np.linalg.eig(v @ v.T)

        idx = D1.argsort()[::-1]
        D1 = D1[idx]
        P1 = P1[:, idx]

        idx = D2.argsort()[::-1]
        D2 = D2[idx]
        P2 = P2[:, idx]

        if not np.isclose(D1[0], D2[0]):
            D2 = -D2
            idx = D2.argsort()[::-1]
            D2 = D2[idx]
            P2 = P2[:, idx]

        a = P1.T
        b = P2.T

        c = np.conj(v).T @ b.T @ a @ u

        left_coset = E @ b.T @ a @ np.conj(E).T
        right_coset = E @ c.T @ np.conj(E).T

        return left_coset, right_coset



from FrozenYoghourt.mode import *


class Maths:

    def mm(*lst, no_times=1):
        lst *= no_times
        return reduce(lambda a, b: a @ b, lst)

    def tp(*lst, no_times=1):
        lst *= no_times
        if Mode.representation == 'numpy':
            return reduce(lambda a, b: np.kron(a, b), lst)
        else:
            return TensorProduct(*lst)

    def svd(A:np.ndarray):
        if Mode.representation == 'numpy':
            return np.linalg.svd(A)
        else:
            AH = A.H
            m, n = A.shape
            if m >= n:
                V, S = (AH * A).diagonalize()

                S_diag = [S[i, i] for i in range(S.rows)]
                ranked = []
                for i, x in enumerate(S.diagonal()):
                    if not x.is_zero:
                        ranked.append(i)

                V = V[:, ranked]
                S = Matrix.diag([sqrt(S[i, i]) for i in range(S.rows) if i in ranked])

                V, _ = V.QRdecomposition()
                U = A * V * S.inv()
            else:
                U, S = (A * AH).diagonalize()

                ranked = []
                for i, x in enumerate(S.diagonal()):
                    if not x.is_zero:
                        ranked.append(i)

                U = U[:, ranked]
                S = Matrix.diag([sqrt(S[i, i]) for i in range(S.rows) if i in ranked])

                U, _ = U.QRdecomposition()
                V = AH * U * S.inv()

            return U, S, V
    
    def to_su(u):
        if Mode.representation == 'numpy':
            return u * complex(np.linalg.det(u)) ** (-1 / np.shape(u)[0])
        else:
            return u * complex(u.det()) ** (-1 / u.shape[0])

    def fast_substitution(matrix:Matrix, variables, values, to_numpy=False):
        substituted_matrix = matrix.subs(list(zip(variables, values)))
        if to_numpy:
            return np.array(substituted_matrix).astype(np.complex)
        else:
            return substituted_matrix

    def optimizing_identities(lhs_fun, rhs_fun, no_var, var_split, random=(0, 2 * np.pi), num_qubits=2, error=1e-7):
        compute_rhs = True
        if var_split == no_var:
            compute_rhs = False

        def cost(x, compute_rhs):
            LHS = lhs_fun(*x[:var_split])

            if compute_rhs:
                RHS = rhs_fun(*x[var_split:])
            else:
                RHS = rhs_fun

            M = (LHS) @ np.conj(RHS).T
            M = M / np.exp(1j * np.angle(M[0, 0]))

            loss = np.linalg.norm(M) - 2 ** (num_qubits / 2)

            return loss

        if random is not None:
            x = np.random.uniform(*random, no_var)
        else:
            x = np.zeros(no_var)

        job = minimize(fun=cost, x0=x, args=(compute_rhs))

        if np.abs(job.fun) < error:
            print("Optimization Sucessful")
        else:
            print("Optimization Unsucessful")

        return job



from FrozenYoghourt.mode import *
from FrozenYoghourt.maths import *


class Gates:

    def Id():
        if Mode.representation == 'numpy':
            return np.array([[1, 0], [0, 1]])
        else:
            return Matrix([[1, 0], [0, 1]])

    def X():
        if Mode.representation == 'numpy':
            return np.array([[0, 1], [1, 0]])
        else:
            return Matrix([[0, 1], [1, 0]])

    def Y():
        if Mode.representation == 'numpy':
            return np.array([[0, -1j], [1j, 0]])
        else:
            return Matrix([[0, -I], [I, 0]])

    def Z():
        if Mode.representation == 'numpy':
            return np.array([[1, 0], [0, -1]])
        else:
            return Matrix([[1, 0], [0, -1]])

    def H():
        if Mode.representation == 'numpy':
            return 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])
        else:
            return 1 / sqrt(2) * Matrix([[1, 1], [1, -1]])

    def Rx(theta:float):
        if Mode.representation == 'numpy':
            return np.array([[np.cos(theta / 2), -1j * np.sin(theta / 2)],
                             [-1j * np.sin(theta / 2), np.cos(theta / 2)]])
        else:
            return Matrix([[cos(theta / 2), -I * sin(theta / 2)],
                           [-I * sin(theta / 2), cos(theta / 2)]])

    def Ry(theta:float):
        if Mode.representation == 'numpy':
            return np.array([[np.cos(theta / 2), -np.sin(theta / 2)],
                             [np.sin(theta / 2), np.cos(theta / 2)]])
        else:
            return Matrix([[cos(theta / 2), -sin(theta / 2)],
                           [sin(theta / 2), cos(theta / 2)]])

    def Rz(theta:float):
        if Mode.representation == 'numpy':
            return np.array([[np.exp(-1j * theta / 2), 0],
                             [0, np.exp(1j * theta / 2)]])
        else:
            return Matrix([[exp(-I * theta / 2), 0],
                           [0, exp(I * theta / 2)]])
        
    def P(theta:float):
        if Mode.representation == 'numpy':
            return np.array([[1, 0], [0, np.exp(1j*theta)]])
        else:
            return Matrix([[1, 0], [0, exp(I*theta)]])

    def U(theta, phi, lam):
        if Mode.representation == 'numpy':
            return np.array([[np.cos(theta / 2), -np.exp(1j * lam) * np.sin(theta / 2)],
                             [np.exp(1j * phi) * np.sin(theta / 2), np.exp(1j * (phi + lam) * np.cos(theta / 2))]])
        else:
            return Matrix([[cos(theta / 2), -exp(I * lam) * sin(theta / 2)],
                           [exp(I * phi) * sin(theta / 2), exp(I * (phi + lam)) * cos(theta / 2)]])

    def CX():
        if Mode.representation == 'numpy':
            return np.array([[1, 0, 0, 0],
                             [0, 1, 0, 0],
                             [0, 0, 0, 1],
                             [0, 0, 1, 0]])
        else:
            return Matrix([[1, 0, 0, 0],
                           [0, 1, 0, 0],
                           [0, 0, 0, 1],
                           [0, 0, 1, 0]])
        
    def CU(control, target, U, no_qubits = 2):
        """
        Manually build the unitary matrix for non-adjacent CU gates
        Parameters:
        -----------
        control: int
            Index of the control qubit (1st qubit is index 0)
        target: int
            Index of the target qubit (1st qubit is index 0)
        U: ndarray
            Target unitary matrix
        edian: bool (True: qiskit convention)
            Qubits order convention
        no_qubits: int
            Number of qubits in the circuit
        Returns:
        --------
        cx_out:
            Unitary matrix for CU gate
        """

        left = [Gates.Id()] * no_qubits
        right = [Gates.Id()] * no_qubits

        if Mode.representation == 'numpy':

            left[control] = np.array([[1, 0], [0, 0]])
            right[control] = np.array([[0, 0], [0, 1]])

        else:

            left[control] = Matrix([[1, 0], [0, 0]])
            right[control] = Matrix([[0, 0], [0, 1]])

        right[target] = U

        cu_out = Maths.tp(*left) + Maths.tp(*right)

        return cu_out

    def Swap():
        if Mode.representation == 'numpy':
            return np.array([[1, 0, 0, 0],
                             [0, 0, 1, 0],
                             [0, 1, 0, 0],
                             [0, 0, 0, 1]])
        else:
            return Matrix([[1, 0, 0, 0],
                           [0, 0, 1, 0],
                           [0, 1, 0, 0],
                           [0, 0, 0, 1]])

    def SQiSW():
        if Mode.representation == 'numpy':
            return np.array([[1, 0, 0, 0],
                             [0, 1 / np.sqrt(2), 1j / np.sqrt(2), 0],
                             [0, 1j / np.sqrt(2), 1 / np.sqrt(2), 0],
                             [0, 0, 0, 1]])
        else:
            return Matrix([[1, 0, 0, 0],
                           [0, 1 / sqrt(2), I / sqrt(2), 0],
                           [0, I / sqrt(2), 1 / sqrt(2), 0],
                           [0, 0, 0, 1]])

    def Magic():
        if Mode.representation == 'numpy':
            return 1 / np.sqrt(2) * np.array([[1, 1j, 0, 0],
                                              [0, 0, 1j, 1],
                                              [0, 0, 1j, -1],
                                              [1, -1j, 0, 0]])
        else:
            return 1 / sqrt(2) * Matrix([[1, I, 0, 0],
                                         [0, 0, I, 1],
                                         [0, 0, I, -1],
                                         [1, -I, 0, 0]])

    def Rxx(theta:float):
        if Mode.representation == 'numpy':
            return np.array([[np.cos(theta / 2), 0, 0, -1j * np.sin(theta / 2)],
                             [0, np.cos(theta / 2), -1j * np.sin(theta / 2), 0],
                             [0, -1j * np.sin(theta / 2), np.cos(theta / 2), 0],
                             [-1j * np.sin(theta / 2), 0, 0, np.cos(theta / 2)]])
        else:
            return Matrix([[cos(theta / 2), 0, 0, -I * sin(theta / 2)],
                           [0, cos(theta / 2), -I * sin(theta / 2), 0],
                           [0, -I * sin(theta / 2), cos(theta / 2), 0],
                           [-I * sin(theta / 2), 0, 0, cos(theta / 2)]])

    def Ryy(theta:float):
        if Mode.representation == 'numpy':
            return np.array([[np.cos(theta / 2), 0, 0, 1j * np.sin(theta / 2)],
                             [0, np.cos(theta / 2), -1j * np.sin(theta / 2), 0],
                             [0, -1j * np.sin(theta / 2), np.cos(theta / 2), 0],
                             [1j * np.sin(theta / 2), 0, 0, np.cos(theta / 2)]])
        else:
            return Matrix([[cos(theta / 2), 0, 0, I * sin(theta / 2)],
                           [0, cos(theta / 2), -I * sin(theta / 2), 0],
                           [0, -I * sin(theta / 2), cos(theta / 2), 0],
                           [I * sin(theta / 2), 0, 0, cos(theta / 2)]])

    def Rzz(theta:float):
        if Mode.representation == 'numpy':
            return np.array([[np.exp(-1j * theta / 2), 0, 0, 0],
                             [0, np.exp(1j * theta / 2), 0, 0],
                             [0, 0, np.exp(1j * theta / 2), 0],
                             [0, 0, 0, np.exp(-1j * theta / 2)]])
        else:
            return Matrix([[exp(-I * theta / 2), 0, 0, 0],
                           [0, exp(I * theta / 2), 0, 0],
                           [0, 0, exp(I * theta / 2), 0],
                           [0, 0, 0, exp(-I * theta / 2)]])
